#!/usr/bin/env python
# Copyright (c) 2024, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.

"""

Datamet
=======

This sub-module contains the Datamet xarray backend for reading Datamet-based radar
data into Xarray structures.

.. autosummary::
   :nosignatures:
   :toctree: generated/

   {}

"""

__all__ = ["DataMetBackendEntrypoint", "open_datamet_datatree"]

__doc__ = __doc__.format("\n   ".join(__all__))

import gzip
import io
import os
import tarfile
from collections import defaultdict
from datetime import datetime, timedelta

import numpy as np
import xarray as xr
from datatree import DataTree
from xarray.backends.common import AbstractDataStore, BackendArray, BackendEntrypoint
from xarray.backends.file_manager import CachingFileManager
from xarray.backends.store import StoreBackendEntrypoint
from xarray.core import indexing
from xarray.core.utils import FrozenDict
from xarray.core.variable import Variable

from ... import util
from ...model import (
    get_altitude_attrs,
    get_azimuth_attrs,
    get_elevation_attrs,
    get_latitude_attrs,
    get_longitude_attrs,
    get_range_attrs,
    get_time_attrs,
    moment_attrs,
    sweep_vars_mapping,
)
from .common import _assign_root, _attach_sweep_groups

#: mapping from DataMet names to CfRadial2/ODIM
datamet_mapping = {
    "UZ": "DBTH",
    "CZ": "DBZH",
    "V": "VRADH",
    "W": "WRADH",
    "ZDR": "ZDR",
    "PHIDP": "PHIDP",
    "RHOHV": "RHOHV",
    "KDP": "KDP",
}


def convert_value(value):
    """
    Try to convert the value to an int or float. If conversion is not possible,
    return the original value.
    """
    try:
        return int(value)
    except (ValueError, TypeError):
        pass

    try:
        return float(value)
    except (ValueError, TypeError):
        pass

    return value


class DataMetFile:
    def __init__(self, filename):
        self.filename = filename
        # Handle .tar.gz case
        if self.filename.endswith(".gz"):
            with gzip.open(self.filename, "rb") as gzip_file:
                tar_bytes = io.BytesIO(gzip_file.read())
            self.tarfile = tarfile.open(fileobj=tar_bytes)
        else:
            self.tarfile = tarfile.open(self.filename, "r")
        self.first_dimension = "azimuth"  # No idea if other scan types are available
        self.scan_metadata = self.get_scan_metadata()
        self.moments = self.scan_metadata["measure"]
        self.data = dict()

    def extract_parameters(self, parameter_path):
        # Extract set of parameters from a file in the tarball
        member = self.tarfile.getmember(parameter_path)
        file = self.tarfile.extractfile(member)
        labels = np.loadtxt(file, delimiter="=", dtype=str, usecols=[0])
        file.seek(0)  # Reset file pointer
        values = np.loadtxt(file, delimiter="=", dtype=str, usecols=[1])
        values = np.array(values, ndmin=1)
        labels = np.array(labels, ndmin=1)
        # Iterate over the labels and values, appending values to the appropriate label key
        parameters = defaultdict(list)
        for label, value in zip(labels, values):
            parameters[label.strip()].append(value.strip())
        parameters = dict(parameters)
        # Convert lists with a single element to individual values
        parameters = {
            k: convert_value(v[0]) if len(v) == 1 else v for k, v in parameters.items()
        }
        return parameters

    def extract_data(self, data_path, dtype):
        # Extract moment data from a file in the tarball
        member = self.tarfile.getmember(data_path + "/SCAN.dat")
        file = self.tarfile.extractfile(member)
        data = np.frombuffer(file.read(), dtype=dtype)
        return data

    def get_scan_metadata(self):
        # Get all metadata at scan level (valid for all sweeps/moments)
        navigation = self.extract_parameters(os.path.join(".", "navigation.txt"))
        archiviation = self.extract_parameters(os.path.join(".", "archiviation.txt"))
        return {**navigation, **archiviation}

    def get_mom_metadata(self, mom, sweep):
        # Get all metadata that is moment and/or sweep dependent
        # Note that DataMet uses 1-indexed but we switch to 0-indexed
        # as is done in other backends
        mom_path = os.path.join(".", mom)
        sweep_path = os.path.join(".", mom, str(sweep + 1))

        generic = self.extract_parameters(os.path.join(sweep_path, "generic.txt"))
        calibration_momlvl = self.extract_parameters(
            os.path.join(mom_path, "calibration.txt")
        )
        calibration_sweeplvl = self.extract_parameters(
            os.path.join(sweep_path, "calibration.txt")
        )
        navigation_var = self.extract_parameters(
            os.path.join(sweep_path, "navigation.txt")
        )
        return {
            **navigation_var,
            **generic,
            **calibration_sweeplvl,
            **calibration_momlvl,
        }

    def get_moment(self, mom, sweep):
        # Get the data for a moment and apply byte to float conversion
        # Note that DataMet uses 1-indexed but we switch to 0-indexed
        # as is done in other backends
        mom_path = os.path.join(".", mom, str(sweep + 1))
        mom_medata = self.get_mom_metadata(mom, sweep)

        bitplanes = 16 if mom == "PHIDP" else int(mom_medata["bitplanes"] or 8)
        nazim = int(mom_medata.get("nlines"))
        nrange = int(mom_medata.get("ncols"))

        dtype = np.uint16 if bitplanes == 16 else np.uint8
        data = self.extract_data(mom_path, dtype)
        data = np.reshape(data, (nazim, nrange))

        return data

    def get_sweep(self, sweep):
        # Get data for all moments and sweeps
        moment_names = self.moments
        if sweep not in self.data:
            self.data[sweep] = dict()

        for moment in moment_names:
            if moment not in self.data[sweep]:
                self.data[sweep][moment] = self.get_moment(moment, sweep)

    def close(self):
        # Clase tarfile
        if self.tarfile is not None:
            self.tarfile.close()


class DataMetArrayWrapper(BackendArray):
    """Wraps array of DataMet RAW data."""

    def __init__(self, datastore, variable_name):
        self.datastore = datastore
        self.group = datastore._group
        self.variable_name = variable_name

        array = self.get_array()
        self.dtype = array.dtype
        self.shape = array.shape

    def _getitem(self, key):
        # read the data and put it into dict
        key = tuple(list(k) if isinstance(k, np.ndarray) else k for k in key)
        array = self.get_array()
        return array[key]

    def __getitem__(self, key):
        return indexing.explicit_indexing_adapter(
            key,
            self.shape,
            indexing.IndexingSupport.BASIC,
            self._getitem,
        )

    def get_array(self):
        ds = self.datastore._acquire()
        ds.get_sweep(self.group)
        return ds.data[self.group][self.variable_name]


class DataMetStore(AbstractDataStore):
    """Store for reading DataMet sweeps."""

    def __init__(self, manager, group=None):
        self._manager = manager
        self._group = int(group[6:])
        self._filename = self.filename
        self._need_time_recalc = False

    @classmethod
    def open(cls, filename, mode="r", group=None, **kwargs):
        manager = CachingFileManager(DataMetFile, filename, kwargs=kwargs)
        return cls(manager, group=group)

    @property
    def filename(self):
        with self._manager.acquire_context(False) as root:
            return root.filename

    @property
    def root(self):
        with self._manager.acquire_context(False) as root:
            return root

    def _acquire(self):
        # Does it need a lock as other backends ?
        ds = self.root
        return ds

    @property
    def ds(self):
        return self._acquire()

    def open_store_variable(self, mom):
        dim = self.root.first_dimension

        data = indexing.LazilyOuterIndexedArray(DataMetArrayWrapper(self, mom))
        encoding = {"group": self._group, "source": self._filename}

        mom_metadata = self.root.get_mom_metadata(mom, self._group)
        add_offset = float(mom_metadata.get("offset") or 0.0)
        scale_factor = float(mom_metadata.get("slope") or 1.0)
        maxval = mom_metadata.get("maxval", None)

        if maxval is not None:
            maxval = float(maxval)
            top = 255
            bottom = float(mom_metadata["bottom"])
            scale_factor = (maxval + add_offset) / (top - bottom)

        mname = datamet_mapping.get(mom, mom)
        mapping = sweep_vars_mapping.get(mname, {})
        attrs = {key: mapping[key] for key in moment_attrs if key in mapping}
        attrs["add_offset"] = add_offset
        attrs["scale_factor"] = scale_factor
        attrs["_FillValue"] = 0
        attrs["coordinates"] = (
            "elevation azimuth range latitude longitude altitude time"
        )
        return {mname: Variable((dim, "range"), data, attrs, encoding)}

    def open_store_coordinates(self, mom):
        scan_mdata = self.root.scan_metadata
        mom_mdata = self.root.get_mom_metadata(mom, self._group)
        dim = self.root.first_dimension

        # Radar coordinates
        lat = scan_mdata["orig_lat"]
        lon = scan_mdata["orig_lon"]
        alt = scan_mdata["orig_alt"]

        rng = (
            mom_mdata["Rangeoff"]
            + np.arange(mom_mdata["ncols"]) * mom_mdata["Rangeres"]
        )
        azimuth = (
            mom_mdata["Azoff"] + np.arange(mom_mdata["nlines"]) * mom_mdata["Azres"]
        )
        # Unravel azimuth
        azimuth[azimuth > 360] -= 360
        elevation = mom_mdata["Eloff"] + np.arange(mom_mdata["nlines"]) * 0

        sweep_mode = "azimuth_surveillance" if dim == "azimuth" else "rhi"
        sweep_number = self._group
        prt_mode = "not_set"
        follow_mode = "not_set"

        raytime = 0  # TODO find out raytime
        raytimes = np.array(
            [
                timedelta(seconds=x * raytime).total_seconds()
                for x in range(azimuth.shape[0] + 1)
            ]
        )
        diff = np.diff(raytimes) / 2.0
        rtime = raytimes[:-1] + diff

        timestr = scan_mdata["dt_acq"]
        time = datetime.strptime(timestr, "%Y-%m-%d-%H%M")

        time_attrs = get_time_attrs(f"{time.isoformat()}Z")

        encoding = {"group": self._group}
        rng = Variable(("range",), rng, get_range_attrs())
        azimuth = Variable((dim,), azimuth, get_azimuth_attrs(), encoding)
        elevation = Variable((dim,), elevation, get_elevation_attrs(), encoding)
        time = Variable((dim,), rtime, time_attrs, encoding)

        coords = {
            "azimuth": azimuth,
            "elevation": elevation,
            "range": rng,
            "time": time,
            "sweep_mode": Variable((), sweep_mode),
            "sweep_number": Variable((), sweep_number),
            "prt_mode": Variable((), prt_mode),
            "follow_mode": Variable((), follow_mode),
            "sweep_fixed_angle": Variable((), float(elevation[0])),
            "longitude": Variable((), lon, get_longitude_attrs()),
            "latitude": Variable((), lat, get_latitude_attrs()),
            "altitude": Variable((), alt, get_altitude_attrs()),
        }

        return coords

    def get_variables(self):
        return FrozenDict(
            (k1, v1)
            for k1, v1 in {
                **{
                    k: v
                    for list_item in [
                        self.open_store_variable(k) for k in self.ds.moments
                    ]
                    for (k, v) in list_item.items()
                },
                **self.open_store_coordinates(self.ds.moments[0]),
            }.items()
        )

    def get_attrs(self):
        return FrozenDict()


class DataMetBackendEntrypoint(BackendEntrypoint):
    """Xarray BackendEntrypoint for DataMet data."""

    description = "Open DataMet files in Xarray"
    url = "https://xradar.rtfd.io/latest/io.html#datamet-data-i-o"

    def open_dataset(
        self,
        filename_or_obj,
        *,
        mask_and_scale=True,
        decode_times=True,
        concat_characters=True,
        decode_coords=True,
        drop_variables=None,
        use_cftime=None,
        decode_timedelta=None,
        group=None,
        reindex_angle=False,
        first_dim="auto",
        site_coords=True,
    ):
        store = DataMetStore.open(
            filename_or_obj,
            group=group,
        )

        store_entrypoint = StoreBackendEntrypoint()

        ds = store_entrypoint.open_dataset(
            store,
            mask_and_scale=mask_and_scale,
            decode_times=decode_times,
            concat_characters=concat_characters,
            decode_coords=decode_coords,
            drop_variables=drop_variables,
            use_cftime=use_cftime,
            decode_timedelta=decode_timedelta,
        )

        # reassign azimuth/elevation/time coordinates
        ds = ds.assign_coords({"azimuth": ds.azimuth})
        ds = ds.assign_coords({"elevation": ds.elevation})
        ds = ds.assign_coords({"time": ds.time})

        ds.encoding["engine"] = "datamet"

        # handle duplicates and reindex
        if decode_coords and reindex_angle is not False:
            ds = ds.pipe(util.remove_duplicate_rays)
            ds = ds.pipe(util.reindex_angle, **reindex_angle)
            ds = ds.pipe(util.ipol_time, **reindex_angle)

        # handling first dimension
        dim0 = "elevation" if ds.sweep_mode.load() == "rhi" else "azimuth"
        if first_dim == "auto":
            if "time" in ds.dims:
                ds = ds.swap_dims({"time": dim0})
            ds = ds.sortby(dim0)
        else:
            if "time" not in ds.dims:
                ds = ds.swap_dims({dim0: "time"})
            ds = ds.sortby("time")

        # assign geo-coords
        if site_coords:
            ds = ds.assign_coords(
                {
                    "latitude": ds.latitude,
                    "longitude": ds.longitude,
                    "altitude": ds.altitude,
                }
            )

        return ds


def open_datamet_datatree(filename_or_obj, **kwargs):
    """Open DataMet dataset as :py:class:`datatree.DataTree`.

    Parameters
    ----------
    filename_or_obj : str, Path, file-like or DataStore
        Strings and Path objects are interpreted as a path to a local or remote
        radar file

    Keyword Arguments
    -----------------
    sweep : int, list of int, optional
        Sweep number(s) to extract, default to first sweep. If None, all sweeps are
        extracted into a list.
    first_dim : str
        Can be ``time`` or ``auto`` first dimension. If set to ``auto``,
        first dimension will be either ``azimuth`` or ``elevation`` depending on
        type of sweep. Defaults to ``auto``.
    reindex_angle : bool or dict
        Defaults to False, no reindexing. Given dict should contain the kwargs to
        reindex_angle. Only invoked if `decode_coord=True`.
    fix_second_angle : bool
        If True, fixes erroneous second angle data. Defaults to ``False``.
    site_coords : bool
        Attach radar site-coordinates to Dataset, defaults to ``True``.
    kwargs : dict
        Additional kwargs are fed to :py:func:`xarray.open_dataset`.

    Returns
    -------
    dtree: datatree.DataTree
        DataTree
    """
    # handle kwargs, extract first_dim
    backend_kwargs = kwargs.pop("backend_kwargs", {})
    # first_dim = backend_kwargs.pop("first_dim", None)
    kwargs["backend_kwargs"] = backend_kwargs

    sweep = kwargs.pop("sweep", None)
    sweeps = []
    kwargs["backend_kwargs"] = backend_kwargs

    if isinstance(sweep, str):
        sweeps = [sweep]
    elif isinstance(sweep, int):
        sweeps = [f"sweep_{sweep}"]
    elif isinstance(sweep, list):
        if isinstance(sweep[0], int):
            sweeps = [f"sweep_{i}" for i in sweep]
        else:
            sweeps.extend(sweep)
    else:
        # Get number of sweeps from data
        dmet = DataMetFile(filename_or_obj)
        sweeps = [
            f"sweep_{i}" for i in range(0, dmet.scan_metadata["elevation_number"])
        ]

    ds = [
        xr.open_dataset(
            filename_or_obj, group=swp, engine=DataMetBackendEntrypoint, **kwargs
        )
        for swp in sweeps
    ]

    ds.insert(0, xr.Dataset())

    # create datatree root node with required data
    dtree = DataTree(data=_assign_root(ds), name="root")
    # return datatree with attached sweep child nodes
    return _attach_sweep_groups(dtree, ds[1:])



#!/usr/bin/env python
# Copyright (c) 2022-2023, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.

"""
XRadar Util
===========

.. autosummary::
   :nosignatures:
   :toctree: generated/

   {}
"""

__all__ = [
    "has_import",
    "get_first_angle",
    "get_second_angle",
    "remove_duplicate_rays",
    "reindex_angle",
    "extract_angle_parameters",
    "ipol_time",
    "rolling_dim",
    "get_sweep_keys",
    "apply_to_sweeps",
]

__doc__ = __doc__.format("\n   ".join(__all__))

import contextlib
import gzip
import importlib.util
import io
import warnings

import datatree as dt
import numpy as np
from scipy import interpolate


def has_import(pkg_name):
    return importlib.util.find_spec(pkg_name)


def get_first_angle(ds):
    """Return name of first angle dimension from given dataset.

    Parameters
    ----------
    ds : xarray.Dataset
        Dateset to get first angle name from.

    Returns
    -------
    first_angle : str
        Name of first angle dimension.
    """
    first_angle = list(set(ds.dims) & set(ds.coords) ^ {"range"})[0]
    if first_angle == "time":
        raise ValueError(
            "first dimension is ``time``, but needed ``azimuth`` or ``elevation``"
        )
    return first_angle


def get_second_angle(ds):
    """Return name of second angle coordinate from given dataset.

    Parameters
    ----------
    ds : xarray.Dataset
        Dateset to get second angle name from.

    Returns
    -------
    out : str
        Name of second angle coordinate.
    """
    return list(
        ((set(ds.coords) | set(ds.variables)) ^ set(ds.dims)) & {"azimuth", "elevation"}
    )[0]


def remove_duplicate_rays(ds):
    """Remove duplicate rays.

    Parameters
    ----------
    ds : xarray.Dataset
        Dateset to remove duplicate rays.

    Returns
    -------
    ds : xarray.Dataset
        Dataset with duplicate rays removed
    """
    first_angle = get_first_angle(ds)

    _, idx = np.unique(ds[first_angle], return_index=True)
    if len(idx) < len(ds[first_angle]):
        # todo:
        # if times have been calculated with wrong number of rays
        # (ODIM_H5, we would need to recalculate the times
        # how should we find out?
        ds = ds.isel({first_angle: idx})

    return ds


def _reindex_angle(ds, array, tolerance, method="nearest"):
    """Reindex first angle.

    Missing values will be filled by variable's _FillValue.

    Parameters
    ----------
    ds : xarray.Dataset
        Dateset to reindex first angle.
    array : array-like
        Array with angle values which the Dataset should reindex to.
    tolerance : float
        Angle tolerance up to which angles should be considered for used method.

    Keyword Arguments
    -----------------
    method : str
        Reindexing method, defaults to "nearest". See :py:meth:`xarray.Dataset.reindex`.

    Returns
    -------
    ds : xarray.Dataset
        Reindexed dataset
    """
    # handle fill value
    fill_value = {
        k: np.asarray(v._FillValue).astype(v.dtype)
        for k, v in ds.items()
        if hasattr(v, "_FillValue")
    }

    angle = get_first_angle(ds)

    # reindex
    ds = ds.reindex(
        {angle: array},
        method=method,
        tolerance=tolerance,
        fill_value=fill_value,
    )
    return ds


def reindex_angle(
    ds,
    start_angle=None,
    stop_angle=None,
    angle_res=None,
    direction=None,
    method="nearest",
    tolerance=None,
):
    """Reindex along first angle.

    Missing values will be filled by variable's ``_FillValue``.

    Parameters
    ----------
    ds : xarray.Dataset
        Dateset to reindex first angle.

    Keyword Arguments
    -----------------
    start_angle : float
        Start angle of dataset.
    stop_angle : float
        Stop angle of dataset.
    angle_res : float
        Angle resolution of the dataset.
    direction : int
        Sweep direction, -1 -> CCW, 1 -> CW.
    method : str
        Reindexing method, defaults to "nearest". See :py:meth:`xarray.Dataset.reindex`.
    tolerance : float
        Angle tolerance up to which angles should be considered for used method.
        Defaults to angle_res / 2.

    Returns
    -------
    ds : xarray.Dataset
        Reindexed dataset
    """
    if tolerance is None:
        tolerance = angle_res / 2.0

    # handle angle order, angle dimension
    first_angle = get_first_angle(ds)
    second_angle = get_second_angle(ds)

    expected_angle_span = abs(stop_angle - start_angle)
    expected_number_rays = int(np.round(expected_angle_span / angle_res, decimals=0))

    # create reindexing angle
    ang = start_angle + direction * np.arange(
        angle_res / 2.0,
        expected_number_rays * angle_res,
        angle_res,
        dtype=ds[first_angle].dtype,
    )

    ds = ds.sortby(first_angle)
    ds = _reindex_angle(ds, ang, tolerance, method=method)

    # check secondary angle coordinate (no nan)
    # set nan values to reasonable median
    sang = ds[second_angle]
    if np.count_nonzero(np.isnan(sang)):
        ds[second_angle] = sang.fillna(sang.median(skipna=True))

    return ds


def extract_angle_parameters(ds):
    """Extract sweep angle parameters from dataset.

    Parameters
    ----------
    ds : xarray.Dataset
        Dateset to reindex first angle.

    Returns
    -------
    angle_dict : dict
        Dictionary with parameters.
    """
    angle_dict = {}

    # 1. get first and second angle (az/el)
    first_angle = get_first_angle(ds)
    second_angle = get_second_angle(ds)
    angle_dict.update(first_angle=first_angle, second_angle=second_angle)

    # 2. keep min/max angle and time
    min_angle = ds[first_angle].min(skipna=True)
    max_angle = ds[first_angle].max(skipna=True)
    min_time = ds.time.min(skipna=True)
    max_time = ds.time.max(skipna=True)
    angle_dict.update(
        min_angle=min_angle.values,
        max_angle=max_angle.values,
        min_time=min_time.values,
        max_time=max_time.values,
    )

    # 3. get unique angles
    unique_angles = np.unique(ds[first_angle])

    # check if azimuth can be sorted
    # check if time can be sorted
    angles_are_unique = len(unique_angles) == ds[first_angle].size
    times_are_unique = len(np.unique(ds.time)) == ds.time.size
    angle_dict.update(
        angles_are_unique=angles_are_unique, times_are_unique=times_are_unique
    )

    # 4. get index and value of first measured angle, aka a1gate
    a1gate_idx = ds.time.argmin(first_angle).values
    a1gate_val = ds.time.idxmin(first_angle).values
    angle_dict.update(a1gate_idx=a1gate_idx, a1gate_val=a1gate_val)

    # 5. angle differences
    fdim = ds[first_angle]
    diff = fdim.diff(first_angle)

    # this captures different angle spacing
    # catches also missing rays and double rays
    # and other erroneous ray alignments which result in different first_angle values

    # 6. this finds the median of those diff-values which are over some quantile
    # in fact this removes angle differences which are too small
    # calculate std/median
    # see GH112, if we get further issues, we might add a warning here in the future
    std = diff.std(skipna=True)
    median = diff.median(skipna=True)
    # remove values above and below std (centered around median), calculate median
    median_diff = diff.where(
        (diff >= (median - std)) & (diff <= (median + std))
    ).median(skipna=True)

    # unique differences
    diffset = set(diff.values)
    # if there are more than 1 unique angle differences that means
    # non-uniform angle spacing
    uniform_angle_spacing = len(diffset) == 1
    angle_dict.update(uniform_angle_spacing=uniform_angle_spacing)

    # 7. ascending/descending/direction
    # PPI ascending CW, descending CCW
    ascending = median_diff > 0
    direction = 1 if ascending else -1
    angle_dict.update(ascending=ascending.values, direction=direction)

    # 8. calculate angle resolution from median_diff and rounding
    angle_res = np.round(median_diff, decimals=2)
    angle_dict.update(angle_res=angle_res.values)

    # 9. start/stop angle
    # so for PPI we assume 0 to 360, for RHI range we try to guess it from the values
    # unfortunately this doesn't work if rays are missing at start or end of the sweep
    # so we need to have the possibility to override
    # also accounts for direction of sweep (CW/CCW, up/down)
    mode = ds.sweep_mode
    if mode == "azimuth_surveillance":
        start_ang = 0
        stop_ang = 360
    elif mode == "rhi":
        # start/stop angle, round to nearest multiple of angle_res
        start_ang = int(np.round(min_angle / angle_res) * angle_res)
        stop_ang = int(np.round(max_angle / angle_res) * angle_res)

    start_ang, stop_ang = (start_ang, stop_ang)[::direction]
    angle_dict.update(start_angle=start_ang, stop_angle=stop_ang)

    # 10. expected angle span
    expected_angle_span = abs(stop_ang - start_ang)
    angle_dict.update(expected_angle_span=expected_angle_span)

    # 11. incomplete sweep
    missing_rays = (angle_res * fdim.size) < expected_angle_span
    excess_rays = (angle_res * fdim.size) > expected_angle_span
    angle_dict.update(missing_rays=missing_rays.values, excess_rays=excess_rays.values)

    # 12. expected number of rays
    expected_number_rays = int(np.round(expected_angle_span / angle_res, decimals=0))
    angle_dict.update(expected_number_rays=expected_number_rays)

    return angle_dict


def _trailing_zeros(da, digits=16, dim=0):
    """Calculate number of trailing zeros for input cast to int."""
    for i in range(digits):
        x = da.dropna(da.dims[dim]).astype(int).values % np.power(10, i)
        if not x.any():
            continue
        else:
            break
    return i - 1


def _ipol_time(da, dim0, a1gate=0, direction=1):
    """Interpolate/extrapolate missing time steps (NaT).

    Parameters
    ----------
    da : xarray.DataArray
        DataArray to interpolate/extrapolate missing timesteps.

    Returns
    -------
    da : xarray.DataArray
        DataArray with interpolated/extrapolated timesteps.
    """
    dtype = da.dtype
    idx = slice(None, None)

    # extract wanted section
    da_sel = da.isel({dim0: idx})

    # get sorting indices along first dimension
    sidx = da_sel[dim0].argsort()

    # special handling for wrap-around angles
    angles = da_sel[dim0].values
    # a1gate should normally only be set for PPI,
    if a1gate > 0:
        angles[-a1gate:] += 360
    da_sel = da_sel.assign_coords({dim0: angles})

    # prepare azimuth array for interpolation
    angles = da_sel[dim0].diff(dim0).cumsum(dim0).pad({dim0: (1, 0)}, constant_values=0)
    da_sel = da_sel.assign_coords({dim0: angles * direction})

    # apply original order for interpolation, get angles
    angles = da_sel.sortby([sidx])[dim0]

    # drop NaT from selection for creation of interpolator
    da_sel = da_sel.dropna(dim0)

    # setup interpolator
    ipol = interpolate.interp1d(
        da_sel[dim0].values,
        da_sel.astype(int),
        fill_value="extrapolate",
        assume_sorted=False,
    )

    # floating point interpolation might introduce spurious digits
    # get least significant digit
    sig = np.power(10, _trailing_zeros(da_sel.time))

    # interpolate and round to the least significant digit
    data = np.rint(ipol(angles) / sig).astype(int) * sig

    # apply interpolated times into original DataArray
    da.loc[{dim0: idx}] = data.astype(dtype)
    return da


def ipol_time(ds, *, a1gate_idx=None, direction=None, **kwargs):
    """Interpolate/extrapolate missing time steps (NaT).

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to interpolate/extrapolate missing timesteps.

    Keyword Arguments
    -----------------
    a1gate_idx : int | None
        First measured gate. 0 assumed, if None.
    direction : int | None
        1: CW, -1: CCW, Clockwise assumed, if None.

    Returns
    -------
    ds : xarray.Dataset
        Dataset with interpolated/extrapolated timesteps.
    """
    # get first dim and sort to get common state
    dim0 = get_first_angle(ds)
    ds = ds.sortby(dim0)

    # return early, if nothing to do
    if not np.isnan(ds.time).any():
        return ds

    if direction is None:
        # set clockwise, rotating in positive direction
        direction = 1

    time = ds.time.astype(int)
    # skip NaT (-9223372036854775808) for amin/amax calculation
    amin = time.where(time > -9223372036854775808).argmin(dim0, skipna=True).values
    amax = time.where(time > -9223372036854775808).argmax(dim0, skipna=True).values
    time = ds.time

    if a1gate_idx is None:
        # if times are not sorted ascending
        if amin > amax:
            # check if we have missing times between amax and amin
            # todo: align with start or end per keyword argument
            if (amin - amax) > 1:
                warnings.warn(
                    "Rays might miss on beginning and/or end of sweep. "
                    "`a1gate` information is needed to fully recover. "
                    "We'll assume sweep start at first valid ray."
                )
            # set a1gate to amin
            a1gate_idx = amin
        else:
            a1gate_idx = 0

    if a1gate_idx > 0:
        # roll first ray to 0-index, interpolate, roll-back
        time = time.roll({dim0: -a1gate_idx}, roll_coords=True)
        time = time.pipe(_ipol_time, dim0, a1gate=a1gate_idx, direction=direction)
        time = time.roll({dim0: a1gate_idx}, roll_coords=True)
    else:
        time = time.pipe(_ipol_time, dim0, direction=direction)

    ds_out = ds.assign({"time": ([dim0], time.values)})
    return ds_out.sortby(dim0)


@contextlib.contextmanager
def _get_data_file(file, file_or_filelike):
    if file_or_filelike == "filelike":
        _open = open
        if file[-3:] == ".gz":
            _open = gzip.open
        with _open(file, mode="r+b") as f:
            yield io.BytesIO(f.read())
    else:
        yield file


def rolling_dim(data, window):
    """Return array with rolling dimension of window-length added at the end."""
    shape = data.shape[:-1] + (data.shape[-1] - window + 1, window)
    strides = data.strides + (data.strides[-1],)
    return np.lib.stride_tricks.as_strided(data, shape=shape, strides=strides)


def get_sweep_keys(dt):
    """Return which nodes in the datatree contain sweep variables

    Parameters
    ----------
    dt : xarray.DataTree
        Datatree to check for sweep_n keys

    Returns
    -------
    keys : list
        List of associated keys with sweep_n
    """
    sweep_group_keys = []
    for key in list(dt.children):
        parts = key.split("_")
        try:
            # Try to set the second part of the tree key to an int
            int(parts[1])
            # Check for "sweep" in the first part of the key
            assert "sweep" in parts[0]
            sweep_group_keys.append(key)

        # This would fail with strings - ex. sweep_group_attrs
        except ValueError:
            pass

        # This would fail if "sweep" not in key - ex. radar_parameters
        except AssertionError:
            pass

    return sweep_group_keys


def apply_to_sweeps(dtree, func, *args, **kwargs):
    """
    Applies a given function to all sweep nodes in the radar volume.

    Parameters
    ----------
    dtree : DataTree
        The DataTree object representing the radar volume.
    func : function
        The function to apply to each sweep.
    *args : tuple
        Additional positional arguments to pass to the function.
    **kwargs : dict
        Additional keyword arguments to pass to the function.

    Returns
    -------
    DataTree
        A new DataTree object with the function applied to all sweeps.
    """
    # Create a new tree dictionary
    tree = {"/": dtree.ds}  # Start with the root Dataset

    # Add all nodes except the root
    tree.update({node.path: node.ds for node in dtree.subtree if node.path != "/"})

    # Apply the function to all sweep nodes and update the tree dictionary
    tree.update(
        {
            node.path: func(dtree[node.path].to_dataset(), *args, **kwargs)
            for node in dtree.match("sweep*").subtree
            if node.path.startswith("/sweep")
        }
    )

    # Return a new DataTree constructed from the modified tree dictionary
    return dt.DataTree.from_dict(tree)


def apply_to_volume(dtree, func, *args, **kwargs):
    """
    Alias for apply_to_sweeps.
    Applies a given function to all sweep nodes in the radar volume.

    Parameters
    ----------
    dtree : DataTree
        The DataTree object representing the radar volume.
    func : function
        The function to apply to each sweep.
    *args : tuple
        Additional positional arguments to pass to the function.
    **kwargs : dict
        Additional keyword arguments to pass to the function.

    Returns
    -------
    DataTree
        A new DataTree object with the function applied to all sweeps.
    """
    return apply_to_sweeps(dtree, func, *args, **kwargs)



#!/usr/bin/env python
# Copyright (c) 2022-2024, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.

"""

ODIM_H5
=======

This sub-module contains the ODIM_H5 xarray backend for reading ODIM_H5-based radar
data into Xarray structures as well as a reader to create a complete datatree.Datatree.

Code ported from wradlib.

Example::

    import xradar as xd
    dtree = xd.io.open_odim_datatree(filename)

.. autosummary::
   :nosignatures:
   :toctree: generated/

   {}

"""

__all__ = [
    "OdimBackendEntrypoint",
    "open_odim_datatree",
]

__doc__ = __doc__.format("\n   ".join(__all__))

import datetime as dt
import io
import warnings

import h5netcdf
import numpy as np
import xarray as xr
from datatree import DataTree
from xarray.backends.common import (
    AbstractDataStore,
    BackendArray,
    BackendEntrypoint,
    find_root_and_group,
)
from xarray.backends.file_manager import CachingFileManager, DummyFileManager
from xarray.backends.locks import SerializableLock, ensure_lock
from xarray.backends.store import StoreBackendEntrypoint
from xarray.core import indexing
from xarray.core.utils import FrozenDict, is_remote_uri
from xarray.core.variable import Variable

from ... import util
from ...model import (
    get_altitude_attrs,
    get_azimuth_attrs,
    get_elevation_attrs,
    get_latitude_attrs,
    get_longitude_attrs,
    get_range_attrs,
    get_time_attrs,
    moment_attrs,
    sweep_vars_mapping,
)
from .common import (
    _assign_root,
    _attach_sweep_groups,
    _fix_angle,
    _get_h5group_names,
    _maybe_decode,
)

HDF5_LOCK = SerializableLock()


def _calculate_angle_res(dim):
    # need to sort dim first
    angle_diff = np.diff(sorted(dim))
    angle_diff2 = np.abs(np.diff(angle_diff))

    # only select angle_diff, where angle_diff2 is less than 0.1 deg
    # Todo: currently 0.05 is working in most cases
    #  make this robust or parameterisable
    angle_diff_wanted = angle_diff[:-1][angle_diff2 < 0.05]
    return np.round(np.nanmean(angle_diff_wanted), decimals=2)


def _get_azimuth_how(how):
    startaz = how["startazA"]
    stopaz = how.get("stopazA", False)
    if stopaz is False:
        # stopazA missing
        # create from startazA
        stopaz = np.roll(startaz, -1)
        stopaz[-1] += 360
    zero_index = np.where(stopaz < startaz)
    stopaz[zero_index[0]] += 360
    azimuth_data = (startaz + stopaz) / 2.0
    azimuth_data[azimuth_data >= 360] -= 360
    return azimuth_data


def _get_azimuth_where(where):
    res = 360.0 / where["nrays"]
    return np.arange(res / 2.0, 360.0, res, dtype="float32")


def _get_fixed_dim_and_angle(where):
    dim = "elevation"

    # try RHI first
    angle_keys = ["az_angle", "azangle"]
    angle = None
    for ak in angle_keys:
        angle = where.get(ak, None)
        if angle is not None:
            break
    if angle is None:
        dim = "azimuth"
        angle = where["elangle"]

    # do not round angle
    # angle = np.round(angle, decimals=1)
    return dim, angle


def _get_elevation_how(how):
    startel = how.get("startelA", False)
    stopel = how.get("stopelA", False)
    if startel is not False and stopel is not False:
        elevation_data = (startel + stopel) / 2.0
    else:
        elevation_data = how["elangles"]
    return elevation_data


def _get_elevation_where(where):
    return np.ones(where["nrays"], dtype="float32") * where["elangle"]


def _get_time_how(how):
    return (how["startazT"] + how["stopazT"]) / 2.0


def _get_time_what(what, where, nrays=None):
    startdate = _maybe_decode(what["startdate"])
    starttime = _maybe_decode(what["starttime"])
    # take care for missing enddate/endtime
    # see https://github.com/wradlib/wradlib/issues/563
    enddate = _maybe_decode(what.get("enddate", what["startdate"]))
    endtime = _maybe_decode(what.get("endtime", what["starttime"]))
    start = dt.datetime.strptime(startdate + starttime, "%Y%m%d%H%M%S")
    end = dt.datetime.strptime(enddate + endtime, "%Y%m%d%H%M%S")
    start = start.replace(tzinfo=dt.timezone.utc).timestamp()
    end = end.replace(tzinfo=dt.timezone.utc).timestamp()
    if nrays is None:
        nrays = where["nrays"]
    if start == end:
        import warnings

        warnings.warn(
            "xradar: Equal ODIM `starttime` and `endtime` "
            "values. Can't determine correct sweep start-, "
            "end- and raytimes.",
            UserWarning,
        )

        time_data = np.ones(nrays) * start
    else:
        delta = (end - start) / nrays
        time_data = np.arange(start + delta / 2.0, end, delta)
        time_data = np.roll(time_data, shift=+where["a1gate"])
    return time_data


def _get_range(where, odim_version=None):
    scale = 1000.0
    if odim_version == "ODIM_H5/V2_4":
        scale = 1.0
    ngates = where["nbins"]
    range_start = where["rstart"] * scale
    bin_range = where["rscale"]
    cent_first = range_start + bin_range / 2.0
    range_data = np.arange(
        cent_first, range_start + bin_range * ngates, bin_range, dtype="float32"
    )
    return range_data, cent_first, bin_range


def _get_time(what, point="start"):
    startdate = _maybe_decode(what[f"{point}date"])
    starttime = _maybe_decode(what[f"{point}time"])
    start = dt.datetime.strptime(startdate + starttime, "%Y%m%d%H%M%S")
    start = np.array(start.replace(tzinfo=dt.timezone.utc)).astype("<M8[s]")
    return start


def _get_a1gate(where):
    return where["a1gate"]


class _H5NetCDFMetadata:
    """Wrapper around OdimH5/Gamic data fileobj for easy access of metadata.

    Parameters
    ----------
    fileobj : file-like
        h5netcdf filehandle.
    group : str
        odim group to acquire

    Returns
    -------
    object : metadata object
    """

    def __init__(self, fileobj, group):
        self._root = fileobj
        self._group = group
        self._how = None
        self._what = None
        self._where = None
        self._dim0 = None
        self._fixed_angle = None

    @property
    def coordinates(self):
        # additional metadata which might come in handy in the future
        # do not forward a1gate and angle_res for now
        # a1gate = self.a1gate
        # angle_res = _calculate_angle_res(locals()[dim])
        # az_attrs["a1gate"] = a1gate
        #
        # if dim == "azimuth":
        #     az_attrs["angle_res"] = angle_res
        # else:
        #     el_attrs["angle_res"] = angle_res

        coordinates = {
            "azimuth": self.azimuth,
            "elevation": self.elevation,
            "time": self.time,
            "range": self.range,
            "sweep_mode": self.sweep_mode,
            "sweep_number": self.sweep_number,
            "prt_mode": self.prt_mode,
            "follow_mode": self.follow_mode,
            "sweep_fixed_angle": self.sweep_fixed_angle,
            "longitude": self.longitude,
            "latitude": self.latitude,
            "altitude": self.altitude,
        }
        return coordinates

    @property
    def first_dim(self):
        warnings.warn(
            "xradar: use of `first_dim` is deprecated, please use `dim0` instead.",
            DeprecationWarning,
        )
        if self._dim0 is None:
            self._dim0, self._fixed_angle = self._get_fixed_dim_and_angle()
        return self._dim0

    @property
    def dim0(self):
        """Get name of first dimension."""
        if self._dim0 is None:
            self._dim0, self._fixed_angle = self._get_fixed_dim_and_angle()
        return self._dim0

    def get_variable_dimensions(self, dims):
        dimensions = []
        for n, _ in enumerate(dims):
            if n == 0:
                dimensions.append(self.dim0)
            elif n == 1:
                dimensions.append("range")
        return tuple(dimensions)

    @property
    def site_coords(self):
        return self._get_site_coords()

    @property
    def how(self):
        """Get how-group attributes."""
        if self._how is None:
            self._how = self.grp["how"].attrs
        return self._how

    @property
    def what(self):
        """Get what-group attributes."""
        if self._what is None:
            self._what = self.grp["what"].attrs
        return self._what

    @property
    def where(self):
        """Get where-group attributes."""
        if self._where is None:
            self._where = self.grp["where"].attrs
        return self._where

    def _get_odim_version(self):
        version = self._root.attrs.get("Conventions", "None")
        return version

    def _get_site_coords(self):
        lon = self._root["where"].attrs["lon"]
        lat = self._root["where"].attrs["lat"]
        alt = self._root["where"].attrs["height"]
        return lon, lat, alt

    @property
    def azimuth(self):
        """Return azimuth Variable."""
        return self._azimuth

    @property
    def elevation(self):
        """Return elevation Variable."""
        return self._elevation

    @property
    def time(self):
        """Return time Variable."""
        return self._time

    @property
    def range(self):
        """Return range Variable."""
        range_data, cent_first, bin_range = self._get_range()
        range_attrs = get_range_attrs(range_data)
        return Variable(("range",), range_data, range_attrs)

    @property
    def longitude(self):
        """Return longitude Variable."""
        return Variable((), self._root["where"].attrs["lon"], get_longitude_attrs())

    @property
    def latitude(self):
        """Return latitude Variable."""
        return Variable((), self._root["where"].attrs["lat"], get_latitude_attrs())

    @property
    def altitude(self):
        """Return altitude Variable."""
        return Variable((), self._root["where"].attrs["height"], get_altitude_attrs())

    @property
    def sweep_fixed_angle(self):
        """Return sweep_fixed_angle Variable."""
        if self._fixed_angle is None:
            self._dim0, self._fixed_angle = self._get_fixed_dim_and_angle()
        return Variable((), self._fixed_angle)

    @property
    def sweep_mode(self):
        """Return sweep_mode Variable."""
        sweep_mode = "azimuth_surveillance" if self.dim0 == "azimuth" else "rhi"
        return Variable((), sweep_mode)

    @property
    def sweep_number(self):
        """Return sweep_number Variable."""
        return Variable((), self._sweep_number)

    @property
    def prt_mode(self):
        """Return prt_mode Variable."""
        return Variable((), "not_set")

    @property
    def follow_mode(self):
        """Return follow_mode Variable."""
        return Variable((), "not_set")


class _OdimH5NetCDFMetadata(_H5NetCDFMetadata):
    """Wrapper around OdimH5 data fileobj for easy access of metadata.

    Parameters
    ----------
    fileobj : file-like
        h5netcdf filehandle.
    group : str
        odim group to acquire

    Returns
    -------
    object : metadata object
    """

    @property
    def ds_what(self):
        return self._get_dset_what()

    @property
    def grp(self):
        return self._root[self._group.split("/")[0]]

    def _get_fixed_dim_and_angle(self):
        return _get_fixed_dim_and_angle(self.where)

    def _get_ray_times(self, nrays=None):
        try:
            time_data = _get_time_how(self.how)
            self._need_time_recalc = False
        except (AttributeError, KeyError, TypeError):
            time_data = _get_time_what(self.what, self.where, nrays)
            self._need_time_recalc = True
        return time_data

    def _get_range(self):
        return _get_range(self.where, odim_version=self._odim_version)

    def _get_time(self, point="start"):
        return _get_time(self.what, point=point)

    def _get_dset_what(self):
        attrs = {}
        what = self._root[self._group]["what"].attrs
        gain = what.get("gain", 1.0)
        offset = what.get("offset", 0.0)
        if not (gain == 1.0 and offset == 0.0):
            attrs["scale_factor"] = gain
            attrs["add_offset"] = offset
        attrs["_FillValue"] = what.get("nodata", None)
        attrs["_Undetect"] = what.get("undetect", 0.0)
        # if no quantity is given, use the group-name
        attrs["quantity"] = _maybe_decode(
            what.get("quantity", self._group.split("/")[-1])
        )
        return attrs

    @property
    def a1gate(self):
        return _get_a1gate(self.where)

    @property
    def _azimuth(self):
        try:
            azimuth = _get_azimuth_how(self.how)
        except (AttributeError, KeyError, TypeError):
            azimuth = _get_azimuth_where(self.where)
        return Variable((self.dim0,), azimuth, get_azimuth_attrs())

    @property
    def _elevation(self):
        try:
            elevation = _get_elevation_how(self.how)
        except (AttributeError, KeyError, TypeError):
            elevation = _get_elevation_where(self.where)
        return Variable((self.dim0,), elevation, get_elevation_attrs())

    @property
    def _time(self):
        return Variable((self.dim0,), self._get_ray_times(), get_time_attrs())

    @property
    def _sweep_number(self):
        """Return sweep number."""
        return int(self._group.split("/")[0][7:]) - 1

    @property
    def _odim_version(self):
        return self._get_odim_version()


class H5NetCDFArrayWrapper(BackendArray):
    """H5NetCDFArrayWrapper

    adapted from https://github.com/pydata/xarray/
    """

    __slots__ = ("datastore", "dtype", "shape", "variable_name")

    def __init__(self, variable_name, datastore):
        self.datastore = datastore
        self.variable_name = variable_name

        array = self.get_array()
        self.shape = array.shape

        dtype = array.dtype
        if dtype is str:
            # use object dtype because that's the only way in numpy to
            # represent variable length strings; it also prevents automatic
            # string concatenation via conventions.decode_cf_variable
            dtype = np.dtype("O")
        self.dtype = dtype

    def __setitem__(self, key, value):
        with self.datastore.lock:
            data = self.get_array(needs_lock=False)
            data[key] = value
            if self.datastore.autoclose:
                self.datastore.close(needs_lock=False)

    def get_array(self, needs_lock=True):
        ds = self.datastore._acquire(needs_lock)
        return ds.variables[self.variable_name]

    def __getitem__(self, key):
        return indexing.explicit_indexing_adapter(
            key, self.shape, indexing.IndexingSupport.OUTER_1VECTOR, self._getitem
        )

    def _getitem(self, key):
        # h5py requires using lists for fancy indexing:
        # https://github.com/h5py/h5py/issues/992
        key = tuple(list(k) if isinstance(k, np.ndarray) else k for k in key)
        with self.datastore.lock:
            array = self.get_array(needs_lock=False)
            return array[key]


def _get_h5netcdf_encoding(self, var):
    """get encoding from h5netcdf Variable

    adapted from https://github.com/pydata/xarray/
    """
    import h5py

    # netCDF4 specific encoding
    encoding = {
        "chunksizes": var.chunks,
        "fletcher32": var.fletcher32,
        "shuffle": var.shuffle,
    }

    # Convert h5py-style compression options to NetCDF4-Python
    # style, if possible
    if var.compression == "gzip":
        encoding["zlib"] = True
        encoding["complevel"] = var.compression_opts
    elif var.compression is not None:
        encoding["compression"] = var.compression
        encoding["compression_opts"] = var.compression_opts

    # save source so __repr__ can detect if it's local or not
    encoding["source"] = self._filename
    encoding["original_shape"] = var.shape

    vlen_dtype = h5py.check_dtype(vlen=var.dtype)
    if vlen_dtype is str:
        encoding["dtype"] = str
    # todo: this might be removed, since xarray is capable from version 2023.X.Y
    elif vlen_dtype is not None:  # pragma: no cover
        # xarray doesn't support writing arbitrary vlen dtypes yet.
        pass
    else:
        encoding["dtype"] = var.dtype
    return encoding


def _get_odim_variable_name_and_attrs(name, attrs):
    if "data" in name:
        name = attrs.pop("quantity")
        # handle non-standard moment names
        try:
            mapping = sweep_vars_mapping[name]
        except KeyError:
            pass
        else:
            attrs.update({key: mapping[key] for key in moment_attrs})
        attrs["coordinates"] = (
            "elevation azimuth range latitude longitude altitude time"
        )
    return name, attrs


class OdimSubStore(AbstractDataStore):
    """Store for reading ODIM data-moments via h5netcdf."""

    def __init__(
        self,
        store,
        group=None,
        lock=False,
    ):
        if not isinstance(store, OdimStore):
            raise TypeError(
                f"Wrong type {type(store)} for parameter store, "
                f"expected 'OdimStore'."
            )

        self._manager = store._manager
        self._group = group
        self._filename = store.filename
        self.is_remote = is_remote_uri(self._filename)
        self.lock = ensure_lock(lock)

    @property
    def root(self):
        with self._manager.acquire_context(False) as root:
            return _OdimH5NetCDFMetadata(root, self._group.lstrip("/"))

    def _acquire(self, needs_lock=True):
        with self._manager.acquire_context(needs_lock) as root:
            ds = root[self._group.lstrip("/")]
        return ds

    @property
    def ds(self):
        return self._acquire()

    def open_store_variable(self, name, var):
        dimensions = self.root.get_variable_dimensions(var.dimensions)
        data = indexing.LazilyOuterIndexedArray(H5NetCDFArrayWrapper(name, self))
        encoding = _get_h5netcdf_encoding(self, var)
        encoding["group"] = self._group
        name, attrs = _get_odim_variable_name_and_attrs(name, self.root.ds_what)

        return name, Variable(dimensions, data, attrs, encoding)

    def open_store_coordinates(self):
        return self.root.coordinates

    def get_variables(self):
        return FrozenDict(
            (k1, v1)
            for k1, v1 in {
                **dict(
                    [
                        self.open_store_variable(k, v)
                        for k, v in self.ds.variables.items()
                    ]
                ),
            }.items()
        )


class OdimStore(AbstractDataStore):
    """Store for reading ODIM dataset groups via h5netcdf."""

    def __init__(self, manager, group=None, lock=False):
        if isinstance(manager, (h5netcdf.File, h5netcdf.Group)):
            if group is None:
                root, group = find_root_and_group(manager)
            else:
                if type(manager) is not h5netcdf.File:
                    raise ValueError(
                        "must supply a h5netcdf.File if the group "
                        "argument is provided"
                    )
                root = manager
            manager = DummyFileManager(root)

        self._manager = manager
        self._group = f"/dataset{int(group[6:])+1}"
        self._filename = self.filename
        self.is_remote = is_remote_uri(self._filename)
        self.lock = ensure_lock(lock)
        self._substore = None
        self._need_time_recalc = False

    @classmethod
    def open(
        cls,
        filename,
        mode="r",
        format=None,
        group=None,
        lock=None,
        invalid_netcdf=None,
        phony_dims=None,
        decode_vlen_strings=True,
    ):
        if isinstance(filename, bytes):
            raise ValueError(
                "can't open netCDF4/HDF5 as bytes "
                "try passing a path or file-like object"
            )

        if format not in [None, "NETCDF4"]:
            raise ValueError("invalid format for h5netcdf backend")

        kwargs = {"invalid_netcdf": invalid_netcdf}
        if phony_dims is not None:
            kwargs["phony_dims"] = phony_dims
        kwargs["decode_vlen_strings"] = decode_vlen_strings

        if lock is None:
            if util.has_import("dask"):
                lock = HDF5_LOCK
            else:
                lock = False

        manager = CachingFileManager(h5netcdf.File, filename, mode=mode, kwargs=kwargs)
        return cls(manager, group=group, lock=lock)

    @property
    def filename(self):
        with self._manager.acquire_context(False) as root:
            return root.filename

    @property
    def substore(self):
        if self._substore is None:
            with self._manager.acquire_context(False) as root:
                subgroups = [
                    "/".join([self._group, k])
                    for k in root[self._group].groups
                    # get data and quality groups
                    if "data" or "quality" in k
                ]
                substore = []
                substore.extend(
                    [
                        OdimSubStore(
                            self,
                            group=group,
                            lock=self.lock,
                        )
                        for group in subgroups
                    ]
                )
                self._substore = substore

        return self._substore

    def open_store_coordinates(self):
        return self.substore[0].open_store_coordinates()

    def get_variables(self):
        return FrozenDict(
            (k1, v1)
            for k1, v1 in {
                **{
                    k: v
                    for substore in self.substore
                    for k, v in substore.get_variables().items()
                },
                **self.open_store_coordinates(),
            }.items()
        )

    def get_attrs(self):
        return FrozenDict()


class OdimBackendEntrypoint(BackendEntrypoint):
    """Xarray BackendEntrypoint for ODIM data.

    Keyword Arguments
    -----------------
    first_dim : str
        Can be ``time`` or ``auto`` first dimension. If set to ``auto``,
        first dimension will be either ``azimuth`` or ``elevation`` depending on
        type of sweep. Defaults to ``auto``.
    reindex_angle : bool or dict
        Defaults to False, no reindexing. Given dict should contain the kwargs to
        reindex_angle. Only invoked if `decode_coord=True`.
    fix_second_angle : bool
        If True, fixes erroneous second angle data. Defaults to ``False``.
    site_coords : bool
        Attach radar site-coordinates to Dataset, defaults to ``True``.
    kwargs : dict
        Additional kwargs are fed to :py:func:`xarray.open_dataset`.
    """

    description = "Open ODIM_H5 (.h5, .hdf5) using h5netcdf in Xarray"
    url = "https://xradar.rtfd.io/en/latest/io.html#odim-h5"

    def open_dataset(
        self,
        filename_or_obj,
        *,
        mask_and_scale=True,
        decode_times=True,
        concat_characters=True,
        decode_coords=True,
        drop_variables=None,
        use_cftime=None,
        decode_timedelta=None,
        format=None,
        group="sweep_0",
        invalid_netcdf=None,
        phony_dims="access",
        decode_vlen_strings=True,
        first_dim="auto",
        reindex_angle=False,
        fix_second_angle=False,
        site_coords=True,
    ):
        if isinstance(filename_or_obj, io.IOBase):
            filename_or_obj.seek(0)

        store = OdimStore.open(
            filename_or_obj,
            format=format,
            group=group,
            invalid_netcdf=invalid_netcdf,
            phony_dims=phony_dims,
            decode_vlen_strings=decode_vlen_strings,
        )

        store_entrypoint = StoreBackendEntrypoint()

        ds = store_entrypoint.open_dataset(
            store,
            mask_and_scale=mask_and_scale,
            decode_times=decode_times,
            concat_characters=concat_characters,
            decode_coords=decode_coords,
            drop_variables=drop_variables,
            use_cftime=use_cftime,
            decode_timedelta=decode_timedelta,
        )

        # reassign azimuth/elevation/time coordinates
        ds = ds.assign_coords({"azimuth": ds.azimuth})
        ds = ds.assign_coords({"elevation": ds.elevation})
        ds = ds.assign_coords({"time": ds.time})

        ds.encoding["engine"] = "odim"

        # handle duplicates and reindex
        if decode_coords and reindex_angle is not False:
            ds = ds.pipe(util.remove_duplicate_rays)
            ds = ds.pipe(util.reindex_angle, **reindex_angle)
            ds = ds.pipe(util.ipol_time, **reindex_angle)

        # handling first dimension
        dim0 = "elevation" if ds.sweep_mode.load() == "rhi" else "azimuth"
        if first_dim == "auto":
            if "time" in ds.dims:
                ds = ds.swap_dims({"time": dim0})
            ds = ds.sortby(dim0)
        else:
            if "time" not in ds.dims:
                ds = ds.swap_dims({dim0: "time"})
            ds = ds.sortby("time")

        # fix second angle
        if fix_second_angle and first_dim == "auto":
            dim1 = {"azimuth": "elevation", "elevation": "azimuth"}[dim0]
            ds = ds.assign_coords({dim1: ds[dim1].pipe(_fix_angle)})

        # assign geo-coords
        ds = ds.assign_coords(
            {
                "latitude": ds.latitude,
                "longitude": ds.longitude,
                "altitude": ds.altitude,
            }
        )

        return ds


def open_odim_datatree(filename_or_obj, **kwargs):
    """Open ODIM_H5 dataset as :py:class:`datatree.DataTree`.

    Parameters
    ----------
    filename_or_obj : str, Path, file-like or DataStore
        Strings and Path objects are interpreted as a path to a local or remote
        radar file

    Keyword Arguments
    -----------------
    sweep : int, list of int, optional
        Sweep number(s) to extract, default to first sweep. If None, all sweeps are
        extracted into a list.
    first_dim : str
        Can be ``time`` or ``auto`` first dimension. If set to ``auto``,
        first dimension will be either ``azimuth`` or ``elevation`` depending on
        type of sweep. Defaults to ``auto``.
    reindex_angle : bool or dict
        Defaults to False, no reindexing. Given dict should contain the kwargs to
        reindex_angle. Only invoked if `decode_coord=True`.
    fix_second_angle : bool
        If True, fixes erroneous second angle data. Defaults to ``False``.
    site_coords : bool
        Attach radar site-coordinates to Dataset, defaults to ``True``.
    kwargs : dict
        Additional kwargs are fed to :py:func:`xarray.open_dataset`.

    Returns
    -------
    dtree: datatree.DataTree
        DataTree
    """
    # handle kwargs, extract first_dim
    backend_kwargs = kwargs.pop("backend_kwargs", {})
    # first_dim = backend_kwargs.pop("first_dim", None)
    sweep = kwargs.pop("sweep", None)
    sweeps = []
    kwargs["backend_kwargs"] = backend_kwargs

    if isinstance(sweep, str):
        sweeps = [sweep]
    elif isinstance(sweep, int):
        sweeps = [f"sweep_{sweep}"]
    elif isinstance(sweep, list):
        if isinstance(sweep[0], int):
            sweeps = [f"sweep_{i+1}" for i in sweep]
        else:
            sweeps.extend(sweep)
    else:
        sweeps = _get_h5group_names(filename_or_obj, "odim")

    ds = [
        xr.open_dataset(filename_or_obj, group=swp, engine="odim", **kwargs)
        for swp in sweeps
    ]

    # todo: apply CfRadial2 group structure below
    ds.insert(0, xr.open_dataset(filename_or_obj, group="/"))

    # create datatree root node with required data
    dtree = DataTree(data=_assign_root(ds), name="root")
    # return datatree with attached sweep child nodes
    return _attach_sweep_groups(dtree, ds[1:])



#!/usr/bin/env python
# Copyright (c) 2023, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.

"""
Projection
^^^^^^^^^^

.. autosummary::
   :nosignatures:
   :toctree: generated/

   {}
"""

__all__ = ["get_earth_radius", "add_crs", "add_crs_tree", "get_crs"]

__doc__ = __doc__.format("\n   ".join(__all__))

import numpy as np
import pyproj
from xarray import Variable


def get_earth_radius(crs, latitude):
    """Return earth radius (in m) for a given Spheroid model (crs) at a given latitude.

    .. math::

        R^2 = \\frac{a^4 \\cos(f)^2 + b^4 \\sin(f)^2}
        {a^2 \\cos(f)^2 + b^2 \\sin(f)^2}

    Parameters
    ----------
    crs : :py:class:`~pyproj.crs.CoordinateSystem`
        spatial reference object
    latitude : float
        geodetic latitude in degrees

    Returns
    -------
    radius : float
        earth radius in meter
    """
    geod = crs.get_geod()
    latitude = np.radians(latitude)
    radius = np.sqrt(
        (
            np.power(geod.a, 4) * np.power(np.cos(latitude), 2)
            + np.power(geod.b, 4) * np.power(np.sin(latitude), 2)
        )
        / (
            np.power(geod.a, 2) * np.power(np.cos(latitude), 2)
            + np.power(geod.b, 2) * np.power(np.sin(latitude), 2)
        )
    )
    return radius


def get_crs(ds, datum="WGS84"):
    """Return :py:class:`pyproj.crs.CoordinateSystem` from ``crs_wkt`` or ``spatial_ref`` coordinate.

    Parameters
    ----------
    ds : xarray.Dataset
    datum : str
        datum string, defaults to 'WGS84'

    Returns
    -------
    proj_crs : :py:class:`~pyproj.crs.CoordinateSystem`
    """
    crs_wkt = (
        ds["crs_wkt"]
        if "crs_wkt" in ds
        else ds["spatial_ref"] if "spatial_ref" in ds else False
    )
    if crs_wkt is not False:
        proj_crs = pyproj.CRS.from_cf(crs_wkt.attrs)
    else:
        proj_crs = pyproj.CRS(
            proj="aeqd",
            datum=datum,
            lon_0=ds.longitude.values,
            lat_0=ds.latitude.values,
        )
    return proj_crs


def add_crs(ds, crs=None, datum="WGS84"):
    """Add ``crs_wkt`` coordinate derived from :py:class:`pyproj.crs.CoordinateSystem`.

    Parameters
    ----------
    ds : xarray.Dataset
    crs : :py:class:`~pyproj.crs.CoordinateSystem`
        ``crs_wkt`` to be added, defaults to ``AEQD`` (with given datum)
    datum : str
        datum string, defaults to 'WGS84'

    Returns
    -------
    ds : xarray.Dataset
        Dataset including ``crs_wkt`` coordinate.
    """
    crs_wkt = Variable((), 0)
    if crs is None:
        proj_crs = get_crs(ds, datum=datum)
    else:
        proj_crs = crs
    crs_wkt.attrs.update(proj_crs.to_cf())
    ds = ds.assign_coords(crs_wkt=crs_wkt)
    return ds


def add_crs_tree(radar, datum="WGS84"):
    """Add ``crs_wkt`` coordinate derived from :py:class:`pyproj.crs.CoordinateSystem`.

    Parameters
    ----------
    radar : xarray.Dataset
    crs : :py:class:`~pyproj.crs.CoordinateSystem`
        ``crs_wkt`` to be added, defaults to ``AEQD`` (with given datum)
    datum : str
        datum string, defaults to 'WGS84'

    Returns
    -------
    radar : datatree.DataTree
        Datatree with sweep datasets including ``crs_wkt`` coordinate.
    """
    for key in list(radar.children):
        if "sweep" in key:
            radar[key].ds = add_crs(radar[key].to_dataset(), datum=datum)
    return radar



#!/usr/bin/env python
# Copyright (c) 2022-2024, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.

"""
XRadar Data Model
=================

The data model for the different xradar DataArrays, Datasets and DataTrees is based on NetCDF4/CfRadial2.1.
It will be aligned to the upcoming WMO standard FM301.

This module contains several helper functions to create minimal DataArrays and Datasets, as well as Datatrees.

Must read:

    * `CfRadial`_.
    * `WMO CF extensions`_.
    * `OPERA/ODIM_H5`_.

Code ported from wradlib.

.. autosummary::
   :nosignatures:
   :toctree: generated/

   {}
"""

__all__ = [
    "create_sweep_dataset",
    "get_azimuth_attrs",
    "get_latitude_attrs",
    "get_longitude_attrs",
    "get_altitude_attrs",
    "get_elevation_attrs",
    "get_moment_attrs",
    "get_range_attrs",
    "get_time_attrs",
    "moment_attrs",
    "sweep_vars_mapping",
    "get_azimuth_dataarray",
    "get_elevation_dataarray",
    "get_range_dataarray",
    "get_sweep_dataarray",
    "get_time_dataarray",
    "required_global_attrs",
    "optional_root_attrs",
    "required_root_vars",
    "optional_root_vars",
    "sweep_coordinate_vars",
    "required_sweep_metadata_vars",
    "optional_sweep_metadata_vars",
    "sweep_dataset_vars",
    "non_standard_sweep_dataset_vars",
    "determine_cfradial2_sweep_variables",
    "conform_cfradial2_sweep_group",
]

__doc__ = __doc__.format("\n   ".join(__all__))

import numpy as np
from xarray import DataArray, Dataset, decode_cf

# This is a temporary setup, since CfRadial2.1 and FM301 are not yet finalized.
# Todo: adhere to standards when they are published

#: required global attributes (root-group)
required_global_attrs = dict(
    [
        ("Conventions", "Cf/Radial"),
        ("version", "Cf/Radial version number"),
        ("title", "short description of file contents"),
        ("instrument_name", "nameThe  of radar or lidar"),
        ("institution", "where the original data were produced"),
        (
            "references",
            "references that describe the data or the methods used to produce it",
        ),
        ("source", "method of production of the original data"),
        ("history", "list of modifications to the original data"),
        ("comment", "miscellaneous information"),
        ("platform_is_mobile", "'true' or 'false', assumed 'false' if missing"),
    ]
)


#: optional global attributes (root-group)
optional_root_attrs = dict(
    [
        ("site_name", "name of site where data were gathered"),
        ("scan_name", "name of scan strategy used, if applicable"),
        ("scan_id", "scan strategy id, if applicable. assumed 0 if missing"),
        (
            "ray_times_increase",
            (
                "'true' or 'false', assumed 'true' if missing. "
                "This is set to true if ray times increase monotonically "
                "throughout all of the sweeps in the volume."
            ),
        ),
        (
            "simulated",
            (
                "'true' or 'false', assumed 'false' if missing. "
                "data in this file are simulated"
            ),
        ),
    ]
)

#: required global variables (root-group)
required_root_vars = {
    "volume_number",
    "time_coverage_start",
    "time_coverage_end",
    "latitude",
    "longitude",
    "altitude",
    "sweep_group_name",
    "sweep_fixed_angle",
}

#: optional global attributes (root-group)
optional_root_vars = dict(
    [
        ("platform_type", None),
        ("instrument_type", None),
        ("primary_axis", None),
        ("altitude_agl", None),
        ("frequency", None),  # cfradial 2.0
        ("status_str", None),  # cfrdadial 2.1
        ("status_xml", "status_str"),  # cfradial 2.0
    ]
)

#: sweep-group coordinate variables
sweep_coordinate_vars = {
    "time",
    "range",
    "frequency",
}

#: required sweep-group metadata variables
required_sweep_metadata_vars = {
    "sweep_number",
    "sweep_mode",
    "follow_mode",
    "prt_mode",
    "sweep_fixed_angle",
    "azimuth",
    "elevation",
}

#: optional sweep-group metadata variables
optional_sweep_metadata_vars = {
    "polarization_mode",
    "polarization_sequence",
    "rays_are_indexed",
    "rays_angle_resolution",
    "qc_procedures",
    "target_scan_rate",
    "scan_rate",
    "antenna_transition",
    "pulse_width",
    "calib_index",
    "rx_range_resolution",
    "prt",
    "prt_ratio",
    "prt_sequence",
    "nyquist_velocity",
    "unambiguous_range",
    "n_samples",
}

#: sweep dataset variable names
sweep_dataset_vars = {
    "DBZH",
    "DBZV",
    "ZH",
    "ZV",
    "DBTH",
    "DBTV",
    "TH",
    "TV",
    "VRADH",
    "VRADV",
    "WRADH",
    "WRADV",
    "ZDR",
    "LDR",
    "LDRH",
    "LDRV",
    "PHIDP",
    "KDP",
    "PHIHX",
    "RHOHV",
    "RHOHX",
    "RHOVX",
    "DBM",
    "DBMHC",
    "DBMHX",
    "DBMVC",
    "DBMVX",
    "SNR",
    "SNRHC",
    "SNRHX",
    "SNRVC",
    "SNRVX",
    "NCP",
    "NCPH",
    "NCPV",
    "RR",
    "REC",
}

#: non-standard sweep dataset variable names
non_standard_sweep_dataset_vars = {
    "DBZ",
    "VEL",
    "VR",
}

# root metadata groups
#: radar_parameters subgroup
radar_parameters_subgroup = dict(
    [
        ("radar_antenna_gain_h", None),
        ("radar_antenna_gain_v", None),
        ("radar_beam_width_h", None),
        ("radar_beam_width_v", None),
        ("radar_receiver_bandwidth", None),  # cfradial2.1
        ("radar_rx_bandwidth", "radar_receiver_bandwidth"),  # cfradial1.X
    ]
)

#: radar_calibration subgroup
radar_calibration_subgroup = dict(
    [
        ("calib_index", None),
        ("time", None),
        ("pulse_width", None),
        ("antenna_gain_h", None),
        ("antenna_gain_v", None),
        ("xmit_power_h", None),
        ("xmit_power_v", None),
        ("two_way_waveguide_loss_h", None),
        ("two_way_waveguide_loss_v", None),
        ("two_way_radome_loss_h", None),
        ("two_way_radome_loss_v", None),
        ("receiver_mismatch_loss", None),
        ("receiver_mismatch_loss_h", None),
        ("receiver_mismatch_loss_v", None),
        ("radar_constant_h", None),
        ("radar_constant_v", None),
        ("probert_jones_correction", None),
        ("dielectric_factor_used", None),
        ("noise_hc", None),
        ("noise_vc", None),
        ("noise_hx", None),
        ("noise_vx", None),
        ("receiver_gain_hc", None),
        ("receiver_gain_vc", None),
        ("receiver_gain_hx", None),
        ("receiver_gain_vx", None),
        ("base_1km_hc", None),
        ("base_dbz_1km_hc", "base_1km_hc"),  # cfradial1
        ("base_1km_vc", None),
        ("base_dbz_1km_vc", "base_1km_vc"),  # cfradial1
        ("base_1km_hx", None),
        ("base_dbz_1km_hx", "base_1km_hx"),  # cfradial1
        ("base_1km_vx", None),
        ("base_dbz_1km_vx", "base_1km_vx"),  # cfradial1
        ("sun_power_hc", None),
        ("sun_power_vc", None),
        ("sun_power_hx", None),
        ("sun_power_vx", None),
        ("noise_source_power_h", None),
        ("noise_source_power_v", None),
        ("power_measure_loss_h", None),
        ("power_measure_loss_v", None),
        ("coupler_forward_loss_h", None),
        ("coupler_forward_loss_v", None),
        ("zdr_correction", None),
        ("ldr_correction_h", None),
        ("ldr_correction_v", None),
        ("system_phidp", None),
        ("test_power_h", None),
        ("test_power_v", None),
        ("receiver_slope_hc", None),
        ("receiver_slope_vc", None),
        ("receiver_slope_hx", None),
        ("receiver_slope_vx", None),
        ("k_squared_water", None),
        ("i0_dbm_hc", None),
        ("i0_dbm_vc", None),
        ("i0_dbm_hx", None),
        ("i0_dbm_vx", None),
        ("dynamic_range_db_hc", None),
        ("dynamic_range_db_vc", None),
        ("dynamic_range_db_hx", None),
        ("dynamic_range_db_vx", None),
        ("dbz_correction", None),
    ]
)

#: georeferencing_correction subgroup
georeferencing_correction_subgroup = dict(
    [
        ("azimuth_correction", None),
        ("elevation_correction", None),
        ("range_correction", None),
        ("longitude_correction", None),
        ("latitude_correction", None),
        ("pressure_altitude_correction", None),
        ("altitude_correction", "radar_altitude_correction"),  # cfradial1
        ("radar_altitude_correction", None),
        (
            "eastward_velocity_correction",
            "eastward_ground_speed_correction",
        ),  # cfradial1
        ("eastward_ground_speed_correction", None),
        (
            "northward_velocity_correction",
            "northward_ground_speed_correction",
        ),  # cfradial1
        ("northward_ground_speed_correction", None),
        ("vertical_velocity_correction", None),
        ("heading_correction", None),
        ("roll_correction", None),
        ("pitch_correction", None),
        ("drift_correction", None),
        ("rotation_correction", None),
        ("tilt_correction", None),
    ]
)

#: required range attributes
range_attrs = {
    "units": "meters",
    "standard_name": "projection_range_coordinate",
    "long_name": "range_to_measurement_volume",
    "axis": "radial_range_coordinate",
    "spacing_is_constant": "true",
    "meters_to_center_of_first_gate": "{}",
    "meters_between_gates": "{}",
}

#: required time attributes
time_attrs = {
    "standard_name": "time",
    "units": "seconds since {}",
}

#: required frequency attributes
frequency_attrs = {
    "standard_name": "",
    "units": "s-1",
}


#: required moment attributes
moment_attrs = {"standard_name", "long_name", "units"}

# todo: align this with sweep_dataset_vars
#: CfRadial 2.1 / FM301 / ODIM_H5 mapping
sweep_vars_mapping = {
    "DBZH": {
        "standard_name": "radar_equivalent_reflectivity_factor_h",
        "long_name": "Equivalent reflectivity factor H",
        "short_name": "DBZH",
        "units": "dBZ",
    },
    "DBZH_CLEAN": {
        "standard_name": "radar_equivalent_reflectivity_factor_h",
        "long_name": "Equivalent reflectivity factor H",
        "short_name": "DBZH_CLEAN",
        "units": "dBZ",
    },
    "DBZV": {
        "standard_name": "radar_equivalent_reflectivity_factor_v",
        "long_name": "Equivalent reflectivity factor V",
        "short_name": "DBZV",
        "units": "dBZ",
    },
    "ZH": {
        "standard_name": "radar_linear_equivalent_reflectivity_factor_h",
        "long_name": "Linear equivalent reflectivity factor H",
        "short_name": "ZH",
        "units": "unitless",
    },
    "ZV": {
        "standard_name": "radar_equivalent_reflectivity_factor_v",
        "long_name": "Linear equivalent reflectivity factor V",
        "short_name": "ZV",
        "units": "unitless",
    },
    "DBZ": {
        "standard_name": "radar_equivalent_reflectivity_factor",
        "long_name": "Equivalent reflectivity factor",
        "short_name": "DBZ",
        "units": "dBZ",
    },
    "DBTH": {
        "standard_name": "radar_equivalent_reflectivity_factor_h",
        "long_name": "Total power H (uncorrected reflectivity)",
        "short_name": "DBTH",
        "units": "dBZ",
    },
    "DBTV": {
        "standard_name": "radar_equivalent_reflectivity_factor_v",
        "long_name": "Total power V (uncorrected reflectivity)",
        "short_name": "DBTV",
        "units": "dBZ",
    },
    "TH": {
        "standard_name": "radar_linear_equivalent_reflectivity_factor_h",
        "long_name": "Linear total power H (uncorrected reflectivity)",
        "short_name": "TH",
        "units": "unitless",
    },
    "TV": {
        "standard_name": "radar_linear_equivalent_reflectivity_factor_v",
        "long_name": "Linear total power V (uncorrected reflectivity)",
        "short_name": "TV",
        "units": "unitless",
    },
    "VRADH": {
        "standard_name": "radial_velocity_of_scatterers_away_from_instrument_h",
        "long_name": "Radial velocity of scatterers away from instrument H",
        "short_name": "VRADH",
        "units": "meters per seconds",
    },
    "VRADV": {
        "standard_name": "radial_velocity_of_scatterers_away_from_instrument_v",
        "long_name": "Radial velocity of scatterers away from instrument V",
        "short_name": "VRADV",
        "units": "meters per second",
    },
    "VR": {
        "standard_name": "radial_velocity_of_scatterers_away_from_instrument",
        "long_name": "Radial velocity of scatterers away from instrument",
        "short_name": "VR",
        "units": "meters per seconds",
    },
    "VRAD": {
        "standard_name": "radial_velocity_of_scatterers_away_from_instrument",
        "long_name": "Radial velocity of scatterers away from instrument",
        "short_name": "VRAD",
        "units": "meters per seconds",
    },
    "VRADDH": {
        "standard_name": "radial_velocity_of_scatterers_away_from_instrument_h",
        "long_name": "Radial velocity of scatterers away from instrument H",
        "short_name": "VRADDH",
        "units": "meters per seconds",
    },
    "WRADH": {
        "standard_name": "radar_doppler_spectrum_width_h",
        "long_name": "Doppler spectrum width H",
        "short_name": "WRADH",
        "units": "meters per seconds",
    },
    "UWRADH": {
        "standard_name": "radar_doppler_spectrum_width_h",
        "long_name": "Doppler spectrum width H",
        "short_name": "UWRADH",
        "units": "meters per seconds",
    },
    "WRADV": {
        "standard_name": "radar_doppler_spectrum_width_v",
        "long_name": "Doppler spectrum width V",
        "short_name": "WRADV",
        "units": "meters per second",
    },
    "WRAD": {
        "standard_name": "radar_doppler_spectrum_width",
        "long_name": "Doppler spectrum width",
        "short_name": "WRAD",
        "units": "meters per second",
    },
    "ZDR": {
        "standard_name": "radar_differential_reflectivity_hv",
        "long_name": "Log differential reflectivity H/V",
        "short_name": "ZDR",
        "units": "dB",
    },
    "UZDR": {
        "standard_name": "radar_differential_reflectivity_hv",
        "long_name": "Log differential reflectivity H/V",
        "short_name": "UZDR",
        "units": "dB",
    },
    "LDR": {
        "standard_name": "radar_linear_depolarization_ratio",
        "long_name": "Log-linear depolarization ratio HV",
        "short_name": "LDR",
        "units": "dB",
    },
    "PHIDP": {
        "standard_name": "radar_differential_phase_hv",
        "long_name": "Differential phase HV",
        "short_name": "PHIDP",
        "units": "degrees",
    },
    "UPHIDP": {
        "standard_name": "radar_differential_phase_hv",
        "long_name": "Differential phase HV",
        "short_name": "UPHIDP",
        "units": "degrees",
    },
    "KDP": {
        "standard_name": "radar_specific_differential_phase_hv",
        "long_name": "Specific differential phase HV",
        "short_name": "KDP",
        "units": "degrees per kilometer",
    },
    "RHOHV": {
        "standard_name": "radar_correlation_coefficient_hv",
        "long_name": "Correlation coefficient HV",
        "short_name": "RHOHV",
        "units": "unitless",
    },
    "URHOHV": {
        "standard_name": "radar_correlation_coefficient_hv",
        "long_name": "Correlation coefficient HV",
        "short_name": "URHOHV",
        "units": "unitless",
    },
    "SNRH": {
        "standard_name": "signal_noise_ratio_h",
        "long_name": "Signal Noise Ratio H",
        "short_name": "SNRH",
        "units": "unitless",
    },
    "SNRV": {
        "standard_name": "signal_noise_ratio_v",
        "long_name": "Signal Noise Ratio V",
        "short_name": "SNRV",
        "units": "unitless",
    },
    "SQIH": {
        "standard_name": "signal_quality_index_h",
        "long_name": "Signal Quality H",
        "short_name": "SQIH",
        "units": "unitless",
    },
    "SQIV": {
        "standard_name": "signal_quality_index_v",
        "long_name": "Signal Quality V",
        "short_name": "SQIV",
        "units": "unitless",
    },
    "CCORH": {
        "standard_name": "clutter_correction_h",
        "long_name": "Clutter Correction H",
        "short_name": "CCORH",
        "units": "unitless",
    },
    "CCORV": {
        "standard_name": "clutter_correction_v",
        "long_name": "Clutter Correction V",
        "short_name": "CCORV",
        "units": "unitless",
    },
    "CMAP": {
        "standard_name": "clutter_map",
        "long_name": "Clutter Map",
        "short_name": "CMAP",
        "units": "unitless",
    },
    "RATE": {
        "standard_name": "rainfall_rate",
        "long_name": "rainfall_rate",
        "short_name": "RATE",
        "units": "mm h-1",
    },
}


def get_longitude_attrs():
    lon_attrs = {
        "long_name": "longitude",
        "units": "degrees_east",
        "standard_name": "longitude",
    }
    return lon_attrs


def get_latitude_attrs():
    lat_attrs = {
        "long_name": "latitude",
        "units": "degrees_north",
        "positive": "up",
        "standard_name": "latitude",
    }
    return lat_attrs


def get_altitude_attrs():
    alt_attrs = {
        "long_name": "altitude",
        "units": "meters",
        "standard_name": "altitude",
    }
    return alt_attrs


def get_range_attrs(rng=None):
    """Get Range CF attributes.

    Parameters
    ----------
    rng : :class:`numpy:numpy.ndarray`
        Array with range values.

    Returns
    -------
    range_attrs : dict
        Dictionary with Range CF attributes.
    """
    range_attrs = {
        "units": "meters",
        "standard_name": "projection_range_coordinate",
        "long_name": "range_to_measurement_volume",
        "axis": "radial_range_coordinate",
    }
    if rng is not None:
        diff = np.diff(rng)
        unique = np.unique(diff)
        if len(unique) == 1:
            spacing = "true"
            range_attrs["meters_between_gates"] = diff[0]
        else:
            spacing = "false"
        range_attrs["spacing_is_constant"] = spacing
        range_attrs["meters_to_center_of_first_gate"] = rng[0]

    return range_attrs


def _calculate_angle_res(dim):
    # need to sort dim first
    angle_diff = np.diff(sorted(dim))
    angle_diff2 = np.abs(np.diff(angle_diff))

    # only select angle_diff, where angle_diff2 is less than 0.1 deg
    # Todo: currently 0.05 is working in most cases
    #  make this robust or parameterisable
    angle_diff_wanted = angle_diff[:-1][angle_diff2 < 0.05]
    return np.round(np.nanmean(angle_diff_wanted), decimals=2)


def get_azimuth_attrs(azi=None):
    """Get Azimuth CF attributes.

    Parameters
    ----------
    azi : :class:`numpy:numpy.ndarray`
        Array with azimuth values

    Returns
    -------
    az_attrs : dict
        Dictionary with Azimuth CF attributes.
    """
    az_attrs = {
        "standard_name": "ray_azimuth_angle",
        "long_name": "azimuth_angle_from_true_north",
        "units": "degrees",
        "axis": "radial_azimuth_coordinate",
    }
    if azi is not None:
        unique = len(np.unique(azi)) == 1
        if not unique:
            a1gate = np.argsort(np.argsort(azi))[0]
            az_attrs["a1gate"] = a1gate
            angle_res = _calculate_angle_res(azi)
            az_attrs["angle_res"] = angle_res
    return az_attrs


def get_elevation_attrs(ele=None):
    """Get Elevation CF attributes.

    Parameters
    ----------
    ele : :class:`numpy:numpy.ndarray`
        Array with elevation values

    Returns
    -------
    el_attrs : dict
        Dictionary with Elevation CF attributes.
    """
    el_attrs = {
        "standard_name": "ray_elevation_angle",
        "long_name": "elevation_angle_from_horizontal_plane",
        "units": "degrees",
        "axis": "radial_elevation_coordinate",
    }
    return el_attrs


def get_time_attrs(date_str="1970-01-01T00:00:00Z", date_unit="seconds"):
    time_attrs = {
        "standard_name": "time",
        "units": f"{date_unit} since {date_str}",
    }
    return time_attrs


def get_moment_attrs(moment):
    """Get Radar Moment CF attributes.

    Parameters
    ----------
    moment : str
        String describing radar moment.

    Returns
    -------
    moment_attrs : dict
        Dictionary with Radar Moment CF attributes.
    """
    return sweep_vars_mapping[moment].copy()


def get_range_dataarray(rng, nbins=None):
    """Create Range DataArray.

    Parameters
    ----------
    rng : float or :class:`numpy:numpy.ndarray`
        range resolution or array with range values.
    nbins: int
        number of bins, must be passed if rng is provided as resolution

    Returns
    -------
    rng_da : :class:`xarray:xarray.DataArray`
        Range DataArray.
    """
    if np.isscalar(rng):
        res2 = rng / 2.0
        rng = np.arange(res2, rng * nbins, rng)
    rng_da = DataArray(data=rng, dims=["range"])
    rng_da.attrs.update(get_range_attrs(rng))
    rng_da.name = "range"
    return rng_da


def get_azimuth_dataarray(azimuth, nrays=None, a1gate=0):
    """Create Azimuth DataArray.

    Parameters
    ----------
    azimuth : float or :class:`numpy:numpy.ndarray`
        Azimuth resolution, fixed azimuth value or array with azimuth values.

    Keyword Arguments
    -----------------
    nrays : int
        number of rays
        If int -> set all rays to same azimuth value.
        If None, given azimuth value is used as resolution.
    a1gate : int
        First measured ray. Defaults to 0.

    Returns
    -------
    azi_da : :class:`xarray:xarray.DataArray`
        Azimuth DataArray.
    """
    if np.isscalar(azimuth):
        if nrays is None:
            res2 = azimuth / 2.0
            nrays = int(360 / azimuth)
            azimuth = np.linspace(res2, 360 - res2, nrays)
        else:
            azimuth = np.full(nrays, azimuth)

    azimuth = np.roll(azimuth, -a1gate)
    azi_da = DataArray(data=azimuth, dims=["time"])
    azi_da.attrs.update(get_azimuth_attrs(azimuth))
    azi_da.name = "azimuth"
    return azi_da


def get_elevation_dataarray(elevation, nrays=None):
    """Create Elevation DataArray.

    Parameters
    ----------
    elevation : float or :class:`numpy:numpy.ndarray`
        elevation resolution or array with elevation values.

    Keyword Arguments
    -----------------
    nrays : int
        If int -> set all rays to same elevation value.
        If None, given elevation value is used as resolution.

    Returns
    -------
    ele_da : :class:`xarray:xarray.DataArray`
        Elevation DataArray.
    """
    if np.isscalar(elevation):
        if nrays is None:
            res2 = elevation / 2.0
            nrays = int(90 / elevation)
            elevation = np.linspace(res2, 90 - res2, nrays)
        else:
            elevation = np.full(nrays, elevation)
    ele_da = DataArray(data=elevation, dims=["time"])
    ele_da.attrs.update(get_elevation_attrs(elevation))
    ele_da.name = "elevation"
    return ele_da


def get_time_dataarray(time, nrays, date_str):
    """Create Time DataArray.

    Parameters
    ----------
    time : float or :class:`numpy:numpy.ndarray`
        time resolution or array with time values ('seconds since'-notation).
    nrays : int
        number of rays
    date_str : str
        String with the form 'YYYY-mm-ddTHH:MM:SS' depicting the reference time.

    Returns
    -------
    time_da : :class:`xarray:xarray.DataArray`
        Time DataArray.
    """
    if np.isscalar(time):
        time = np.arange(0, nrays * time - time / 2, time)
    time_attrs = {
        "standard_name": "time",
        "units": f"seconds since {date_str}",
    }
    time_da = DataArray(data=time, dims=["time"])
    time_da.attrs.update(time_attrs)
    time_da.name = "time"
    return time_da


def get_sweep_dataarray(data, moment, fill=None):
    """Create Sweep Moment DataArray.

    Parameters
    ----------
    data : tuple or :class:`numpy:numpy.ndarray`
        tuple with shape of DataArray or 2d-array.
    moment : str
        String describing radar moment.

    Keyword Arguments
    -----------------
    fill : float
        Fill Value for new created array. Only used if data is tuple.

    Returns
    -------
    swp_da : :class:`xarray:xarray.DataArray`
        Sweep Moment DataArray.
    """
    if isinstance(data, tuple):
        if fill is None:
            data = np.tile(np.arange(data[1]), (data[0], 1))
        else:
            data = np.full(data, fill)
    swp_da = DataArray(data=data, dims=["time", "range"])
    swp_da.attrs.update(get_moment_attrs(moment))
    swp_da.name = moment
    return swp_da


def create_sweep_dataset(**kwargs):
    """Create Sweep Dataset with Coordinates.

    This function creates a simple Dataset with all needed coordinates.

    Radar moment variables can be assigned to this Dataset.

    Keyword Arguments
    -----------------
    shape : tuple
        tuple with shape of DataArray
    time : float or :class:`numpy:numpy.ndarray`
        time resolution or array with time values ('seconds since'-notation).
    date_str : str
        String with the form 'YYYY-mm-ddTHH:MM:SS' depicting the reference time.
    rng : float or :class:`numpy:numpy.ndarray`
        range resolution or array with range values.
    azimuth : float or :class:`numpy:numpy.ndarray`
        azimuth resolution or array with azimuth values.
    direction : int
        1 - CW/UP, -1 CCW/DOWN
    a1gate : int
        First measured ray. Defaults to 0. Only used for PPI.
    elevation : float or :class:`numpy:numpy.ndarray`
        elevation resolution or array with elevation values.
    sweep : str
        "PPI" or "RHI".

    Returns
    -------
    swp_ds : :class:`xarray:xarray.Dataset`
        Sweep Dataset.
    """
    shape = kwargs.pop("shape", None)
    sweep = kwargs.pop("sweep", "PPI")

    # set default shape for nrays, nbins
    if shape is None:
        nbins = 1000
        # default resolution/values (1deg azimuth, 1deg elevation)
        if sweep == "PPI":
            azimuth = kwargs.pop("azimuth", 1.0)
            elevation = kwargs.pop("elevation", 1.0)
        else:
            azimuth = kwargs.pop("azimuth", 1.0)
            elevation = kwargs.pop("elevation", 1.0)
    else:
        nbins = shape[1]
        azimuth = kwargs.pop("azimuth", None)
        elevation = kwargs.pop("elevation", None)
        if azimuth is not None and elevation is not None:
            raise ValueError(
                "Either `shape` or `azimuth` and `elevation` can be specified."
            )
        if sweep == "PPI":
            if elevation is None:
                raise ValueError("elevation need to be specified for PPI sweep")
            if azimuth is not None:
                raise ValueError("given azimuth value will be ignored for PPI sweep")
            azimuth = 360 / shape[0]
        else:
            if azimuth is None:
                raise ValueError("azimuth need to be specified for RHI sweep")
            if elevation is not None:
                raise ValueError("given elevation value will be ignored for RHI sweep")
            elevation = 90 / shape[0]

    a1gate = kwargs.pop("a1gate", 0)
    time = kwargs.pop("time", 0.25)
    date_str = kwargs.pop("date_str", "2022-08-27T10:00:00")
    rng = kwargs.pop("rng", 100)
    direction = kwargs.pop("direction", 1)

    # calculate according to PPI/RHI
    if sweep == "PPI":
        azimuth = get_azimuth_dataarray(azimuth, nrays=None, a1gate=a1gate)[::direction]
        nrays = azimuth.shape[0]
        elevation = get_elevation_dataarray(elevation, nrays=nrays)
    else:
        elevation = get_elevation_dataarray(elevation, nrays=None)
        nrays = elevation.shape[0]
        azimuth = get_azimuth_dataarray(azimuth, nrays=nrays, a1gate=a1gate)

    # get time array using retrieved number of rays
    time = get_time_dataarray(time, nrays, date_str)
    rng = get_range_dataarray(rng, nbins)

    # get site coordinates, Locarno Monti, MeteoSwiss
    # trivia: the place where xradar was born in August 2022
    altitude = 375
    latitude = 46.172541
    longitude = 8.7877271

    ds = Dataset(
        coords=dict(
            time=time,
            range=rng,
            azimuth=azimuth,
            elevation=elevation,
            longitude=([], longitude, get_longitude_attrs()),
            latitude=([], latitude, get_latitude_attrs()),
            altitude=([], altitude, get_altitude_attrs()),
        )
    )

    return decode_cf(ds)


def determine_cfradial2_sweep_variables(obj, optional, dim0):
    # "calculate" variables to keep
    keep_vars = set()
    # mandatory coordinates
    keep_vars |= sweep_coordinate_vars
    # required metadata
    keep_vars |= required_sweep_metadata_vars
    # all moment fields
    # todo: strip off non-conforming
    # this also handles cfradial1 n_points layout
    keep_vars |= {
        k
        for k, v in obj.data_vars.items()
        if any(d in v.dims for d in ["range", "n_points"])
    }
    # optional variables
    if optional:
        keep_vars |= {k for k, v in obj.data_vars.items() if dim0 in v.dims}
    return keep_vars


def conform_cfradial2_sweep_group(obj, optional, dim0=None):
    if dim0 is None:
        # handling first dimension
        dim0 = "elevation" if obj.sweep_mode.load().astype(str) == "rhi" else "azimuth"
        if dim0 not in obj.dims:
            dim0 = "time"
            assert dim0 in obj.dims

    keep_vars = determine_cfradial2_sweep_variables(obj, optional, dim0)
    # calculate variables to remove and remove them
    var = set(obj.data_vars)
    var |= set(obj.coords)
    remove_vars = var ^ keep_vars
    # only remove variables if in dataset
    remove_vars &= var
    out = obj.drop_vars(remove_vars)
    out.attrs = {}

    # swap dims, if needed
    if dim0 != "time" and dim0 in obj.dims:
        out = out.swap_dims({dim0: "time"})
    # sort in any case
    out = out.sortby("time")

    return out



#!/usr/bin/env python
# Copyright (c) 2022, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.

"""

CfRadial2 output
================

This sub-module contains the writer for export of CfRadial2-based radar
data.

Code ported from wradlib.

Example::

    import xradar as xd
    dtree = xd.io.to_cfradial2(dtree, filename)

.. autosummary::
   :nosignatures:
   :toctree: generated/

   {}

"""

__all__ = [
    "to_cfradial2",
]

__doc__ = __doc__.format("\n   ".join(__all__))

from importlib.metadata import version

from datatree import DataTree

from ...model import conform_cfradial2_sweep_group
from ...util import has_import


def to_cfradial2(dtree, filename, engine=None, timestep=None):
    """Save DataTree to CfRadial2 compliant file.

    Parameters
    ----------
    dtree : datatree.DataTree
        DataTree with CfRadial2 groups.
    filename : str
        output filename

    Keyword Arguments
    ----------------
    timestep : int
        timestep of wanted volume, currently not used
    engine : str
        Either `netcdf4` or `h5netcdf`.
    """
    if engine is None:
        if has_import("netCDF4"):
            engine == "netcdf4"
        elif has_import("h5netcdf"):
            engine == "h5netcdf"
        else:
            raise ImportError(
                "xradar: ``netCDF4`` or ``h5netcdf`` needed to perform this operation."
            )

    # iterate over DataTree and make subgroups cfradial2 compliant
    for grp in dtree.groups:
        if "sweep" in grp:
            dtree[grp] = DataTree(
                conform_cfradial2_sweep_group(
                    dtree[grp].to_dataset(),
                    optional=False,
                )
            )

    root = dtree["/"].to_dataset()
    # fix Conventions
    root.attrs["Conventions"] = "Cf/Radial"
    root.attrs["version"] = "2.0"
    # add xradar version to history
    xradar_version = version("xradar")
    root.attrs["history"] += f": xradar v{xradar_version} CfRadial2 export"

    # write DataTree
    dtree.to_netcdf(filename, engine=engine)



#!/usr/bin/env python
# Copyright (c) 2022-2023, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.

import numpy as np
import pytest
from numpy.testing import assert_almost_equal

import xradar
from xradar.georeference import antenna_to_cartesian, get_x_y_z


def test_antenna_to_cartesian():
    ranges = np.arange(0, 1000, 100)
    azimuths = np.arange(0, 300, 30)
    elevations = np.arange(0, 50, 5)

    # Apply georeferencing to this sample data
    x, y, z = antenna_to_cartesian(ranges, azimuths, elevations, site_altitude=375.0)

    # Check to see if the origin contains all 0s
    assert_almost_equal(x[0], 0)
    assert_almost_equal(y[0], 0)
    assert_almost_equal(z[0], 375.0)

    # Make sure that at 180 degrees, x is (close to) 0
    assert_almost_equal(x[np.where(azimuths == 180.0)], 0)

    # Make sure that at 270 degrees, y is (close to) 0
    assert_almost_equal(y[np.where(azimuths == 270.0)], 0)


def test_get_x_y_z():
    # Create default xradar dataset
    ds = xradar.model.create_sweep_dataset()

    # apply georeferencing
    ds = get_x_y_z(ds)

    # Check to see if the mean of the first range bins are 0
    # we have ranges at center of bin
    origin = ds.isel(range=0).reset_coords().mean("time")
    assert_almost_equal(origin.x, 0)
    assert_almost_equal(origin.y, 0)
    # center of first bin has already some added elevation
    assert_almost_equal(origin.z, ds.altitude + 0.87276752)

    # Make sure that at 180 degrees, x is (close to) 0
    # select the two beams around 180deg and calculate mean
    sel = ds.where((ds.azimuth >= 179) & (ds.azimuth <= 181), drop=True)
    sel = sel.reset_coords().isel(range=0).mean("time")
    np.testing.assert_almost_equal(sel.x, 0)
    np.testing.assert_approx_equal(sel.y, -50, significant=3)

    # Make sure that at 270 degrees, y is (close to) 0
    # select the two beams around 270deg and calculate mean
    sel = ds.where((ds.azimuth >= 269) & (ds.azimuth <= 271), drop=True)
    sel = sel.reset_coords().isel(range=0).mean("time")
    np.testing.assert_almost_equal(sel.y, 0)
    np.testing.assert_approx_equal(sel.x, -50, significant=3)

    # Make sure spatial_ref has been added with the correct values
    assert ds.crs_wkt == 0
    crs = {
        "semi_major_axis": 6378137.0,
        "semi_minor_axis": 6356752.314245179,
        "inverse_flattening": 298.257223563,
        "reference_ellipsoid_name": "WGS 84",
        "longitude_of_prime_meridian": 0.0,
        "prime_meridian_name": "Greenwich",
        "geographic_crs_name": "unknown",
        "horizontal_datum_name": "World Geodetic System 1984",
        "projected_crs_name": "unknown",
        "grid_mapping_name": "azimuthal_equidistant",
        "latitude_of_projection_origin": 46.172541,
        "longitude_of_projection_origin": 8.7877271,
        "false_easting": 0.0,
        "false_northing": 0.0,
    }
    for key, value in crs.items():
        if isinstance(value, float):
            assert ds.crs_wkt.attrs[key] == pytest.approx(value)
        else:
            assert ds.crs_wkt.attrs[key] == value
